/* Low level interface to registers/memory for SunOS kernel debugging.
   Copyright (C) 1986, 1987 Free Software Foundation, Inc.
   Copyright (C) 1992,1993,1994 by John Heidemann <johnh@cs.ucla.edu>

   This file is derived from rem-multi/remote_inflow.c from gdb 4.5.

	@(#)kgdb_inflow.c	1.10 (John Heidemann) 9/12/94

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

#include <sys/param.h>
#include <signal.h>
#include <sys/ptrace.h>
#include <machine/debug/allregs.h>
#include <debug/debugger.h>

#include "defs.h"
#include "wait.h"
#ifndef WRETCODE
#define WRETCODE(w) WEXITSTATUS(w)
#endif

#include "kgdb_gdb.h"



/***************Begin MY defs*********************/
char registers[REGISTER_BYTES];

/* Index within `registers' of the first byte of the space for
   register N.  */


char buf2[MAX_REGISTER_RAW_SIZE];
/***************End MY defs*********************/

#ifdef NEW_SUN_PTRACE
#include <sys/ptrace.h>
#include <machine/reg.h>
#endif

extern char **environ;
extern int errno;
extern int inferior_pid;
int query();
CORE_ADDR read_register();

/* Nonzero if we are debugging an attached outside process
   rather than an inferior.  */

void
print_some_registers(regs)
	int regs[];
{
	register int i;
	for (i = 0; i < 18; i++) {
		printf("reg[%d] = %x\n", i, regs[i]);
	}
}



/* Kill the inferior process.  Make us have no inferior.  */

void
kgdb_kill_inferior ()
{
	/*
	 * Real work happens in close_remote.
	 */
}

int
kgdb_wait(status)
	int *status;
{
	int trap;
	extern struct allregs regsave;

	postspinner();
	doswitch();
	prespinner();

	/*
	 * We might use this, but it will always
	 * be either 0xfe or 0xff.
	 */
	trap = (regsave.r_tbr & 0xff0) >> 4;

	*status = ((SIGTRAP) << 8);

	return inferior_pid;
}


/* Resume execution of the inferior process.
   If STEP is nonzero, single-step it.
   If SIGNAL is nonzero, give it that signal.  */

unsigned char
kgdb_resume (step, signal,status)
	int step;
	int signal;
	char *status;
{
	int pid;
	WAITTYPE w;

	errno = 0;
	switch (step) {
	case 1:  /* single step */
		dotrace = 1;
		/*FALLTHROUGH*/
	case 0:  /* continue */
		dorun = 1;
		break;
	case 2:   /* just do the wait */
		break;
	};
	pid = kgdb_wait(&w);

	*status = 'S';
	return WSTOPSIG(w);
}

int
verify_wim(cwp, wim)
	int cwp, wim;
{
	int i;
	for (i=cwp; i>0; i--)
		wim >>= 1;
	return (wim & 1) == 0;
}

int
find_prevcwp(cwp)
	int cwp;
{
	extern int nwindows;
	/*
	 * +1 because the register window stack builds down
	 * and we want the previous window
	 */
	return (cwp+nwindows+1) % nwindows;
}
int
find_nextcwp(cwp)
	int cwp;
{
	extern int nwindows;
	return (cwp+nwindows-1) % nwindows;
}

#define CWPMASK 0x1f
#define extract_cwp(PSR) (((PSR) & CWPMASK))


void
kgdb_fetch_inferior_registers (regno)
	int regno;
{
	int cwp, wim, nextcwp;
	extern struct allregs regsave;
	/* extern int nwindows; */
	int realpsr;

#define BELIEVE_STATIC   /* don't bother with bzeros */
	if (regno != -1)
		panic ("kgdb_fetch_inferior_registers: regno!=-1");

	/*
	 * now get the regs out of regsave
	 *
	 * The plan:
	 * First: g0-7 o0-7 l0-7 and i0-7
	 * Next: f0-31
	 * Then: y psr wim tbr pc npc fpsr cpsr
	 *
	 * But: there's a hitch.  The regs are raw, just saved by
	 * locore.s:_trap.  This means that the things done
	 * when taking a trap (see Sparc Risc User's Guide
	 * Section 2.7.5 by Cypress Semiconductor) have been done.
	 * Here we undo them.
	 * In particular, we need to (1) put the S bit back,
	 * (2) increment the cwp ("going to the previous cwp").
	 * We don't do (1) because gdb doesn't care and it
	 * should always be set anyway when debugging the kernel.
	 */
	realpsr = regsave.r_psr;
	cwp = find_prevcwp(extract_cwp(regsave.r_psr));
	realpsr = (realpsr & ~CWPMASK) | cwp;
	wim = regsave.r_wim;

	/*
	 * First: g0-7 o0-7 l0-7 and i0-7
	 */
#ifndef BELIEVE_STATIC
	bzero(&registers[REGISTER_BYTE(G0_REGNUM)], KGDB_REGSIZE);
#endif
	bcopy(regsave.r_globals, &registers[REGISTER_BYTE(G1_REGNUM)], 7*KGDB_REGSIZE);
	if (verify_wim(cwp, wim)) {
		nextcwp = find_nextcwp(cwp);
#if 0
		printf ("kgdb_fetch_inferior_registers: wim=%d cwp=%d, next=%d, nwindows=%d\n", wim, cwp, nextcwp, nwindows);
		{ int i;
		  for (i=0; i<nwindows; i++)
			  printf ("window[%d] i0=%x\n", i, regsave.r_window[i].rw_in[0]);
	  	}
#endif
		bcopy(regsave.r_window[cwp].rw_local,
		      &registers[REGISTER_BYTE(L0_REGNUM)],
		      8*KGDB_REGSIZE);
		bcopy(regsave.r_window[cwp].rw_in,
		      &registers[REGISTER_BYTE(I0_REGNUM)],
		      8*KGDB_REGSIZE);
		bcopy(regsave.r_window[nextcwp].rw_in,
		      &registers[REGISTER_BYTE(O0_REGNUM)],
		      8*KGDB_REGSIZE);
	} else {
		printf ("wim %x says cwp %d is invalid.\n", regsave.r_wim, cwp);
		bzero(&registers[REGISTER_BYTE(O0_REGNUM)], 24*KGDB_REGSIZE);
	};
	/*
	 * Next: f0-31
	 * fp registers are in fpregs,
	 * but unfortunately we can't get at that
	 * from C.  Ignore this fact for now.
	 */
#ifndef BELEIVE_STATIC
	bzero(&registers[REGISTER_BYTE(FP0_REGNUM)], 32*KGDB_REGSIZE);
#endif
	/*
	 * Then: y psr wim tbr pc npc fps cps
	 */
	bcopy(&regsave.r_y, &registers[REGISTER_BYTE(Y_REGNUM)], KGDB_REGSIZE);
	bcopy(&realpsr, &registers[REGISTER_BYTE(PS_REGNUM)], KGDB_REGSIZE);
	bcopy(&regsave.r_wim, &registers[REGISTER_BYTE(WIM_REGNUM)], KGDB_REGSIZE);
	bcopy(&regsave.r_tbr, &registers[REGISTER_BYTE(TBR_REGNUM)], KGDB_REGSIZE);
	bcopy(&regsave.r_pc, &registers[REGISTER_BYTE(PC_REGNUM)], KGDB_REGSIZE);
	bcopy(&regsave.r_npc, &registers[REGISTER_BYTE(NPC_REGNUM)], KGDB_REGSIZE);
#if 1
	/* regsave lacks these fields */
#ifndef BELEIVE_STATIC
	bzero(&registers[REGISTER_BYTE(FPS_REGNUM)], KGDB_REGSIZE);
	bzero(&registers[REGISTER_BYTE(CPS_REGNUM)], KGDB_REGSIZE);
#endif
#else
	bcopy(&regsave.r_fps, &registers[REGISTER_BYTE(FPS_REGNUM)], KGDB_REGSIZE);
	bcopy(&regsave.r_cps, &registers[REGISTER_BYTE(CPS_REGNUM)], KGDB_REGSIZE);
#endif

}

/* Store our register values back into the inferior.
   If REGNO is -1, do this for all registers.
   Otherwise, REGNO specifies which register (so we can save time).  */

int
kgdb_store_inferior_registers (regno)
	int regno;
{
	int userpsr, kernelpsr;
	int cwp, wim, nextcwp;
	extern struct allregs regsave;
	/* extern int nwindows; */

	if (regno != -1)
		panic ("kgdb_store_inferior_registers: regno!=-1\n");

	/*
	 * now put the regs into regsave
	 *
	 * The plan:
	 * First: g0-7 o0-7 l0-7 and i0-7
	 * Next: f0-31
	 * Then: y psr wim tbr pc npc fpsr cpsr
	 *
	 * Remember to undo the hacks we made to the psr.
	 */

	/*
	 * First: g0-7 o0-7 l0-7 and i0-7
	 */
	bcopy(&registers[REGISTER_BYTE(G1_REGNUM)], regsave.r_globals, 7*KGDB_REGSIZE);
	bcopy (&registers[REGISTER_BYTE(PS_REGNUM)], &userpsr, KGDB_REGSIZE);
	cwp = find_nextcwp(extract_cwp(userpsr));
	kernelpsr = (userpsr & ~CWPMASK) | (cwp & CWPMASK);
	if (kernelpsr != regsave.r_psr) {
		printf ("kgdb kgdb_store_inferior_registers: WARNING, %x regsave.r_psr != adjusted registers[PSR] %x\n", regsave.r_psr, kernelpsr);
	};
	wim = regsave.r_wim;

	if (verify_wim(cwp, wim)) {
		nextcwp = find_nextcwp(cwp);
		bcopy(&registers[REGISTER_BYTE(L0_REGNUM)],
		      regsave.r_window[cwp].rw_local,
		      8*KGDB_REGSIZE);
		bcopy(&registers[REGISTER_BYTE(I0_REGNUM)],
		      regsave.r_window[cwp].rw_in,
		      8*KGDB_REGSIZE);
		bcopy(&registers[REGISTER_BYTE(O0_REGNUM)],
		      regsave.r_window[nextcwp].rw_in,
		      8*KGDB_REGSIZE);
	} else {
		printf ("wim %x says cwp %d is invalid.\n", regsave.r_wim, cwp);
	};
	/*
	 * Next: f0-31
	 * NEEDSWORK: Where are those fp regs?  Do they matter?
	 */
#ifndef BELEIVE_STATIC
	bzero(&registers[REGISTER_BYTE(FP0_REGNUM)], 32*KGDB_REGSIZE);
#endif
	/*
	 * Then: y psr wim tbr pc npc fps cps
	 */
	bcopy(&registers[REGISTER_BYTE(Y_REGNUM)], &regsave.r_y, KGDB_REGSIZE);
	/* bcopy(&kernelpsr, &regsave.r_psr, KGDB_REGSIZE); -- proved equal above */
	bcopy(&registers[REGISTER_BYTE(WIM_REGNUM)], &regsave.r_wim, KGDB_REGSIZE);
	bcopy(&registers[REGISTER_BYTE(TBR_REGNUM)], &regsave.r_tbr, KGDB_REGSIZE);
	bcopy(&registers[REGISTER_BYTE(PC_REGNUM)], &regsave.r_pc, KGDB_REGSIZE);
	bcopy(&registers[REGISTER_BYTE(NPC_REGNUM)], &regsave.r_npc, KGDB_REGSIZE);
#if 1
	/* regsave lacks these fields */
#ifndef BELEIVE_STATIC
	bzero(&registers[REGISTER_BYTE(FPS_REGNUM)], KGDB_REGSIZE);
	bzero(&registers[REGISTER_BYTE(CPS_REGNUM)], KGDB_REGSIZE);
#endif
#else
	bcopy(&regsave.r_fps, &registers[REGISTER_BYTE(FPS_REGNUM)], KGDB_REGSIZE);
	bcopy(&regsave.r_cps, &registers[REGISTER_BYTE(CPS_REGNUM)], KGDB_REGSIZE);
#endif
	return 0;
}



/* Copy LEN bytes from inferior's memory starting at MEMADDR
   to debugger memory starting at MYADDR.  */

void
kgdb_read_inferior_memory (memaddr, myaddr, len)
     CORE_ADDR memaddr;
     char *myaddr;
     int len;
{
	register int i;
	/* Round starting address down to longword boundary.  */
	register CORE_ADDR addr = memaddr & - sizeof (int);
	/* Round ending address up; get number of longwords that makes.  */
	register int count
		= (((memaddr + len) - addr) + sizeof (int) - 1) / sizeof (int);
	/* Allocate buffer of that many longwords.  */
	register int *buffer = (int *) alloca (count * sizeof (int));

	/* Read all the longwords */
	for (i = 0; i < count; i++, addr += sizeof (int)) {
		buffer[i] = ptrace (PTRACE_PEEKTEXT, inferior_pid, addr, 0);
	}

	/* Copy appropriate bytes out of the buffer.  */
	scopy ((char *) buffer + (memaddr & (sizeof (int) - 1)), myaddr, len);
}

/* Copy LEN bytes of data from debugger memory at MYADDR
   to inferior's memory at MEMADDR.
   On failure (cannot write the inferior)
   returns the value of errno.  */

int
kgdb_write_inferior_memory (memaddr, myaddr, len)
	CORE_ADDR memaddr;
	char *myaddr;
	int len;
{
	register int i;
	/* Round starting address down to longword boundary.  */
	register CORE_ADDR addr = memaddr & - sizeof (int);
	/* Round ending address up; get number of longwords that makes.  */
	register int count
		= (((memaddr + len) - addr) + sizeof (int) - 1) / sizeof (int);
	/* Allocate buffer of that many longwords.  */
	register int *buffer = (int *) alloca (count * sizeof (int));
	extern int errno;

	/* Fill start and end extra bytes of buffer with existing memory data.  */

	buffer[0] = ptrace (PTRACE_PEEKTEXT, inferior_pid, addr, 0);

	if (count > 1) {
		buffer[count - 1]
			= ptrace (PTRACE_PEEKTEXT, inferior_pid,
				  addr + (count - 1) * sizeof (int), 0);
	}

	/* Copy data to be written over corresponding part of buffer */

	bcopy (myaddr, (char *) buffer + (memaddr & (sizeof (int) - 1)), len);

	/* Write the entire buffer.  */

	for (i = 0; i < count; i++, addr += sizeof (int)) {
		errno = 0;
		ptrace (PTRACE_POKETEXT, inferior_pid, addr, buffer[i]);
		if (errno)
			return errno;
	}

	return 0;
}


